
#region Header

/* /////////////////////////////////////////////////////////////////////////////////////////////////
Morph Editor for Virt-A-Mate by Fabio Den.

Written in Visual Studio
	
Version 1.0 2021-07-06
	Initial release.

License: CC BY-SA
	
///////////////////////////////////////////////////////////////////////////////////////////////// */

using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using UnityEngine.UI;
using UnityEngine.Events;
using UnityEngine.Animations;
using System.Linq;
using MVR.FileManagementSecure;
using SimpleJSON;

#endregion

public class MorphEditorFree : MVRScript
{

    #region Reference

    private void Reference()
    {
		string filtSelect = filterSelect.value.val;
		string filtEdit = filterEdit.value.val;

		foreach (Slot slot in slotList)
		{
			// checkbox visibility
			slot.SetVisible(false, true);

			// slider visibility
			slot.SetVisible(true, true);

			// checkbox value
			bool selected = slot.GetSelected();
			slot.SetSelected(true);

			// slider value
			float value = slot.GetValue();
			slot.SetValue(value);
		}
	}

	#endregion

	#region Serialize values

	public void LoadMorphs()
	{
		try
		{
			SuperController.singleton.GetMediaPathDialog(ImportFileSelected2, _saveExt, _saveFolder);
		}
		catch (Exception exc)
		{
			SuperController.LogError($"MorphEditor: Failed to open file dialog: {exc}");
		}
	}

	private void ImportFileSelected2(string path)
	{
		if (string.IsNullOrEmpty(path)) return;

		try
		{
			JSONClass jc = JSON.Parse(SuperController.singleton.ReadFileIntoString(path)) as JSONClass;
			this.StartCoroutine(LoadExportedJSON2(jc));
		}
		catch (Exception exc)
		{
			SuperController.LogError($"MorphEditor: Failed to import morph values: {exc}");
		}
	}

	private IEnumerator LoadExportedJSON2(JSONClass jc)
	{
		if (jc.HasKey("morphValueCount") == false)
		{
			SuperController.LogMessage("Invalid MorphEditor File!");
			yield return 0;
		}

		filterEdit.value.val = "";

		foreach (Slot slot in slotList)
		{
			slot.SetSelected(false);
		}

		int count = int.Parse(jc["morphValueCount"].Value);

		for (int j = 1; j <= count; j++)
		{
			string name = jc["MorphName" + j.ToString()].Value;
			float value = float.Parse(jc["MorphValue" + j.ToString()]);

			Slot myslot = null;

			foreach (Slot slot in slotList)
			{
				if (slot.morph.displayName.Equals(name))
				{
					myslot = slot;
					break;
				}
			}

			DAZMorph morph = morphControl.GetMorphByDisplayName(name.Replace("!", ""));

			if (morph != null)
			{
				morph.morphValue = value;
				if (myslot == null) 
				{ 
					myslot = AddSlot(morph, false);
					if (myslot != null)
					{ 
						selectedCount++;
						myslot.origValue = 0F;
					}
				}

				if (myslot != null)
					myslot.SetSelected(true);
			}
		}

		yield return new WaitForSecondsRealtime(0.1f);

		SelectMode(true);
		updateStatus();
	}

	public void SaveMorphs()
	{
		try
		{
			var fileBrowserUI = SuperController.singleton.fileBrowserUI;
			fileBrowserUI.SetTitle("Save morph values");
			fileBrowserUI.fileRemovePrefix = null;
			fileBrowserUI.hideExtension = false;
			fileBrowserUI.keepOpen = false;
			fileBrowserUI.fileFormat = _saveExt;
			fileBrowserUI.defaultPath = _saveFolder;
			fileBrowserUI.showDirs = true;
			fileBrowserUI.shortCuts = null;
			fileBrowserUI.browseVarFilesAsDirectories = false;
			fileBrowserUI.SetTextEntry(true);
			fileBrowserUI.Show(ExportFileSelected2);
			fileBrowserUI.ActivateFileNameField();
		}
		catch (Exception exc)
		{
			SuperController.LogError($"MorphEditor: Failed to save file dialog: {exc}");
		}
	}

	private void ExportFileSelected2(string path)
	{
		if (string.IsNullOrEmpty(path)) return;
		if (!path.ToLower().EndsWith($".{_saveExt}")) path += $".{_saveExt}";

		try
		{
			JSONClass jc = new JSONClass();
			// jc = SuperController.singleton.GetSaveJSON(null, true, false);
			FillExportJSON2(jc);

			SuperController.singleton.SaveJSON(jc, path);
			SuperController.singleton.DoSaveScreenshot(path);
		}
		catch (Exception exc)
		{
			SuperController.LogError($"MorphEditor: Failed to export morph values: {exc}");
		}
	}

	public void FillExportJSON2(JSONClass jc)
	{
		int count = 0;

		foreach (Slot slot in slotList)
		{
			if (slot.GetSelected())
				count++;
		}

		jc.Add("morphValueCount", count.ToString());

		int counter = 0;

		foreach (Slot slot in slotList)
		{
			if (slot.GetSelected())
			{
				counter++;
				jc.Add("MorphName" + counter.ToString(), slot.morph.displayName);
				jc.Add("MorphValue" + counter.ToString(), slot.GetValue().ToString());
			}
		}
	}

	#endregion

	#region Serialize selection

	public void LoadSelection() 
	{
		try
		{
			SuperController.singleton.GetMediaPathDialog(ImportFileSelected, _saveExt, _saveFolderSel);
		}
		catch (Exception exc)
		{
			SuperController.LogError($"MorphEditor: Failed to open file dialog: {exc}");
		}
	}

	private void ImportFileSelected(string path)
	{
		if (string.IsNullOrEmpty(path)) return;

		SelectNone();

		try
		{
			JSONClass jc = JSON.Parse(SuperController.singleton.ReadFileIntoString(path)) as JSONClass;
			this.StartCoroutine(LoadExportedJSON(jc));
		}
		catch (Exception exc)
		{
			SuperController.LogError($"MorphEditor: Failed to import selection: {exc}");
		}
	}

	private IEnumerator LoadExportedJSON(JSONClass jc)
	{
		if (jc.HasKey("morphCount") == false)
		{
			SuperController.LogMessage("Invalid MorphEditor File!");
			yield return 0;
		}

		filterSelect.value.val = "";

		foreach (Slot slot in slotList)
		{
			slot.SetSelected(false);
		}

		int count = int.Parse(jc["morphCount"].Value);

		for (int j = 1; j <= count; j++)
		{
			string name = jc["Morph" + j.ToString()].Value;
			Slot myslot = null;

			foreach (Slot slot in slotList)
			{
				if (slot.morph.displayName.Equals(name))
				{
					myslot = slot;
					break;
				}
			}

			DAZMorph morph = morphControl.GetMorphByDisplayName(name.Replace("!", ""));

			if (morph != null)
			{
				if (myslot == null)
				{ 
					myslot = AddSlot(morph, false);
					//if (myslot != null) { selectedCount++; }
				}

				if (myslot != null)
					myslot.SetSelected(true);
			}
		}

		yield return new WaitForSecondsRealtime(0.1f);

		SelectMode(false);
		updateStatus();
	}

	public void SaveSelection()
	{
		SuperController.singleton.ShowMonitorUI();

		try
		{
			var fileBrowserUI = SuperController.singleton.fileBrowserUI;
			fileBrowserUI.SetTitle("Save current morph selection");
			fileBrowserUI.fileRemovePrefix = null;
			fileBrowserUI.hideExtension = false;
			fileBrowserUI.keepOpen = false;
			fileBrowserUI.fileFormat = _saveExt;
			fileBrowserUI.defaultPath = _saveFolderSel;
			fileBrowserUI.showDirs = true;
			fileBrowserUI.shortCuts = null;
			fileBrowserUI.browseVarFilesAsDirectories = false;
			fileBrowserUI.SetTextEntry(true);
			fileBrowserUI.Show(ExportFileSelected);
			fileBrowserUI.ActivateFileNameField();
		}
		catch (Exception exc)
		{
			SuperController.LogError($"MorphEditor: Failed to save file dialog: {exc}");
		}
	}

	private void ExportFileSelected(string path)
	{
		if (string.IsNullOrEmpty(path)) return;
		if (!path.ToLower().EndsWith($".{_saveExt}")) path += $".{_saveExt}";

		try
		{
			JSONClass jc = new JSONClass();
			// jc = SuperController.singleton.GetSaveJSON(null, true, false);
			FillExportJSON(jc);

			SuperController.singleton.SaveJSON(jc, path);
			SuperController.singleton.DoSaveScreenshot(path);
		}
		catch (Exception exc)
		{
			SuperController.LogError($"MorphEditor: Failed to export selection: {exc}");
		}
	}

	public void FillExportJSON(JSONClass jc)
	{
		int count = 0;

		foreach (Slot slot in slotList)
		{
			if (slot.GetSelected())
				count++;
		}

		jc.Add("morphCount", count.ToString());

		int counter = 0;

		foreach (Slot slot in slotList)
		{
			if (slot.GetSelected())
			{
				counter++;
				jc.Add("Morph" + counter.ToString(), slot.morph.displayName);
			}
		}
	}

	#endregion 

	#region Selection

	public void SelectNone()
	{
		foreach (Slot slot in slotList)
		{
			// checkbox value
			if (slot.checkVisible)
				slot.SetSelected(false);
		}
	}

	public void InverseSelection()
	{
		foreach (Slot slot in slotList)
		{
			// checkbox value
			if (slot.checkVisible)
				slot.SetSelected(!slot.GetSelected());
		}
	}

	public void SelectAll()
	{
		foreach (Slot slot in slotList)
		{
			// checkbox value
			if (slot.checkVisible)
				slot.SetSelected(true);
		}
	}

	#endregion

	#region values

	public void OriginalValues()
	{
		foreach (Slot slot in slotList)
		{
			// slider value
			if (slot.GetSelected())
				slot.SetValue(slot.GetOriginalValue());
		}
	}

	public void ZeroValues()
	{
		foreach (Slot slot in slotList)
		{
			// slider value
			if (slot.GetSelected())
				slot.SetValue(0F);
		}
	}
	#endregion

	#region Filter

	private void ShowAll()
    {
		filterSelect.value.val = "All";
		FilterSelect();
    }

	private void ShowSelected()
	{
		filterSelect.value.val = "Selected";
		FilterSelect();
	}

	public void FilterSelect()
	{
		string filtSelect = filterSelect.value.val.ToLower();

		if (filtSelect.Equals("all"))
        {
			filterEdit.value.val = "";

			foreach (Slot slot in slotList)
			{
				// checkbox visibility
				slot.SetVisible(false, true);

				// slider visibility
				slot.SetVisible(true, false);
			}
		}
		else if (filtSelect.Equals("selected"))
        {
			foreach (Slot slot in slotList)
			{
				// checkbox visibility
				slot.SetVisible(false, slot.GetSelected());

				// slider visibility
				slot.SetVisible(true, false);
			}
		}
		else
        {
			if (filtSelect.Length > 0)
			{
				filterEdit.value.val = "";

				foreach (Slot slot in slotList)
				{
					// checkbox visibility
					slot.SetVisible(false, slot.morph.displayName.ToLower().Contains(filtSelect)); //  || slot.GetSelected()

					// slider visibility
					slot.SetVisible(true, false);
				}
			}
			else
			{
				foreach (Slot slot in slotList)
				{
					// checkbox visibility
					slot.SetVisible(false, slot.GetSelected());

					// slider visibility
					slot.SetVisible(true, false);
				}
			}
		}
	}

	public void FilterEdit()
	{
		string filtEdit = filterEdit.value.val.ToLower();

		if (filtEdit.Equals("all"))
		{
			filterSelect.value.val = "";

			foreach (Slot slot in slotList)
			{
				// slider visibility
				slot.SetVisible(true, true);

				// checkbox value
				slot.SetSelected(true);

				// checkbox visibility
				slot.SetVisible(false, false);
			}
		}
		else
		{
			if (filtEdit.Length > 0)
			{
				filterSelect.value.val = "";

				foreach (Slot slot in slotList)
				{
					// slider visibility
					slot.SetVisible(true, slot.morph.displayName.ToLower().Contains(filtEdit));

					// checkbox value
					slot.SetSelected(slot.morph.displayName.ToLower().Contains(filtEdit));

					// checkbox visibility
					slot.SetVisible(false, false);
				}
			}
			else
			{
				foreach (Slot slot in slotList)
				{
					bool value = slot.GetSelected();

					// slider visibility
					slot.SetVisible(true, value);

					// checkbox visibility
					slot.SetVisible(false, false);
				}
			}
		}
	}

	#endregion 

	#region Variables

	private float sep = 5.0F; 
	public bool editing = false;
	private LabelEx status;
	private GenerateDAZMorphsControlUI morphControl = null;
	private List<DAZMorph> allmorphs = null;
	private List<DAZMorph> actmorphs = new List<DAZMorph>();
	private List<Slot> slotList = new List<Slot>();
	private int selectedCount = 0;
	public int Counter = 0;
	public bool sideToggle;
	public bool sideSlider;
	private bool firstTime = true;
	private const string _saveExt = "json";
	private const string _saveFolder = "Custom/Assets/fabio/MorphEditor/";
	private const string _saveFolderSel = "Custom/Assets/fabio/MorphSelector/";

	// select
	private ButtonEx loadSelbtn;
	private ButtonEx saveSelbtn;
	private ButtonEx selectNonbtn;
	private ButtonEx syncbtn;
	private ButtonEx selectAllbtn;
	private LabelEx filterSelect;
	private ButtonEx filterSelbtn;
	private ButtonEx showAllbtn;
	private ButtonEx showSelbtn;

	// edit
	private ButtonEx loadMorphbtn;
	private ButtonEx saveMorphbtn;
	private ButtonEx origValbtn;
	private ButtonEx zeroValbtn;
	private SpacerEx sp3;
	private SpacerEx sp4;
	private SpacerEx sp5 = null;
	private LabelEx filterEdit;
	private ButtonEx filterEditbtn;

	#endregion

	#region Morph Functions

	private void GetActiveMorphs()
	{
		foreach (DAZMorph morph in allmorphs)
		{
			if (myApproximation(morph.morphValue, 0.0f, 0.000001f) == false)
			{
				bool exist = false;

				foreach (Slot slot in slotList)
				{
					if (slot.morph.displayName.Equals(morph.displayName))
					{
						exist = true;
						break;
					}
				}

				if (exist == false) { actmorphs.Add(morph); }
			}
		}

		actmorphs.Sort(CompareMorphs);
	}

	public int CompareMorphs(DAZMorph x, DAZMorph y)
	{
		if (x == null || y == null)
		{
			return 0;
		}

		// "CompareTo()" method
		return x.displayName.CompareTo(y.displayName);
	}

	private bool myApproximation(float a, float b, float tolerance)
	{
		return (Mathf.Abs(a - b) < tolerance);
	}

	#endregion

	#region Slot

	public class Slot
	{
		private MorphEditorFree parent;
		private ToggleEx toggle;
		private SliderEx slider;
		private bool isbold;
		public DAZMorph morph;
		public float origValue;
		public bool checkVisible;

		public void SetSelected(bool selected)
		{
			toggle.value.val = selected;
		}

		public bool GetSelected()
		{
			return toggle.value.val;
		}

		public void SetValue(float value)
		{
			slider.value.val = value;
			morph.morphValue = value;
		}

		public float GetValue()
		{
			return morph.morphValue;
		}

		public float GetMinValue()
		{
			return morph.min;
		}

		public float GetMaxValue()
		{
			return morph.max;
		}

		public float GetOriginalValue()
		{
			return origValue;
		}

		public void SetVisible(bool Is_slider, bool visible)
		{
			if (Is_slider)
            {
				if (visible)
					slider.Show();
				else
					slider.Hide();
            }	
			else
            {
				if (visible)
                {
					checkVisible = true;
					toggle.Show();
				}
				else
                {
					checkVisible = false;
					toggle.Hide();
				}
			}
		}

		public Slot(DAZMorph morph, MorphEditorFree parent, bool isbold)
		{
			this.morph = morph;
			this.parent = parent;
			this.isbold = isbold;
			origValue = morph.morphValue;
			CreateSlot();
		}

		private void CreateSlot()
		{
			parent.sideToggle = !parent.sideToggle;
			parent.sideSlider = !parent.sideSlider;

			// create checkbox
			string name = morph.displayName;

			if (isbold)
			{
				// name = "<b>" + name[0].ToString() + "</b>" + name.Substring(1);
				name = "<b>" + name + "</b>";
			}

			toggle = parent.GUICreateCheckbox(parent, "Check" + parent.Counter.ToString(), name, parent.editing, parent.sideToggle, SyncCheckBox);
			slider = parent.GUICreateSlider(parent, "Slider" + parent.Counter.ToString(), morph.displayName, GetValue(), parent.sideSlider, GetMinValue(), GetMaxValue(), OnValueChanged);

			SetVisible(false, !parent.editing);
			SetVisible(true, parent.editing);	
		}

		private void SyncCheckBox(JSONStorableBool flag)
		{
			toggle.value.val = flag.val;
			UpdateStatus(flag.val);
		}

		private void OnValueChanged(float value)
		{
			morph.morphValue = value;
		}

		private void UpdateStatus(bool selected)
		{
			if (selected)
				parent.selectedCount++;
			else
				parent.selectedCount--;

			if (parent.selectedCount < 0)
				parent.selectedCount = 0;

			if (parent.selectedCount > parent.actmorphs.Count)
				parent.selectedCount = parent.actmorphs.Count;

			parent.updateStatus();
		}

		public void Remove()
		{
			parent.RemoveToggle(toggle.toggle);
			parent.RemoveSlider(slider.slider);
		}
	}

	#endregion

	#region Mode

	public void Mode1()
	{
		filterEdit.value.val = "";
		SelectMode(false);
	}

	public void Mode2()
	{
		SelectMode(true);
	}

	public void SelectMode(bool edit)
	{
		editing = edit;

		if (sp5 != null)
		{
			RemoveSpacer(sp5.spacer);
		}

		if (editing)
		{
			loadSelbtn.Hide();
			saveSelbtn.Hide();
			selectNonbtn.Hide();
			selectAllbtn.Hide();
			filterSelect.Hide();
			filterSelbtn.Hide();
			showAllbtn.Hide();
			showSelbtn.Hide();

			filterEditbtn.Show();
			filterEdit.Show();
			loadMorphbtn.Show();
			saveMorphbtn.Show();
			origValbtn.Show();
			zeroValbtn.Show();
			sp3.Hide();
			sp4.Hide();
		}
		else
		{
			loadSelbtn.Show();
			saveSelbtn.Show();
			selectNonbtn.Show();
			selectAllbtn.Show();
			filterSelect.Show();
			filterSelbtn.Show();
			showAllbtn.Show();
			showSelbtn.Show();

			filterEditbtn.Hide();
			filterEdit.Hide();
			loadMorphbtn.Hide();
			saveMorphbtn.Hide();
			origValbtn.Hide();
			zeroValbtn.Hide();
			sp3.Hide();
			sp4.Hide();
		}

		if (editing)
        {
			FilterEdit();
        }
		else
        {
			FilterSelect();
		}

		sp5 = GUICreateSpacer(this, true, 800F);
	}

	#endregion

	#region Events

	public override void Init()
	{
		Install();

		morphControl = (containingAtom.GetStorableByID("geometry") as DAZCharacterSelector)?.morphsControlUI;
		allmorphs = morphControl.GetMorphs();
		GetActiveMorphs();

		SetupGUI();

		Synchronize(false);
	}

	void Start()
	{
		foreach (Slot slot in slotList)
		{
			if (slot.morph.displayName.Equals("Body Size")) { slot.SetSelected(true); }
			if (slot.morph.displayName.Equals("Breast diameter")) { slot.SetSelected(true); }
			if (slot.morph.displayName.Equals("Breath1")) { slot.SetSelected(true); }
			if (slot.morph.displayName.Equals("Eyes Round")) { slot.SetSelected(true); }
			if (slot.morph.displayName.Equals("Face thin")) { slot.SetSelected(true); }
			if (slot.morph.displayName.Equals("Jaw Size")) { slot.SetSelected(true); }
			if (slot.morph.displayName.Equals("Mouth Width")) { slot.SetSelected(true); }
			if (slot.morph.displayName.Equals("Young Breast1")) { slot.SetSelected(true); }
		}

		// initial mode
		SelectMode(true);

		updateStatus();
		filterSelect.value.val = "Selected";
		// quickList[0].saveClick();
	}

	public void OnEnable()
	{

	}

	public void OnDisable()
	{
		
	}

	public void OnDestroy()
	{
		foreach (Slot slot in slotList)
		{
			slot.Remove();
		}

		slotList.Clear();
	}

	#endregion

	#region Button colors

	public class BClr
	{
		public static Color sync = new Color32(12, 95, 158, 255);
		public static Color load = new Color32(152, 12, 99, 255);
		public static Color mode = sync;
		public static Color cont = sync;
		public static Color quck = new Color32(161, 72, 140, 255);

		public static void Paint(ButtonEx btn, Color back, Color text)
        {
			btn.button.buttonColor = back;
			btn.button.textColor = text;
        }
	}

	#endregion

	#region GUI

		private void SetupGUI()
	{
		// screen
		CreateSpacer(false).height = sep;
		CreateSpacer(true).height = sep;
		status = GUICreateLabel(this, " ", 116.0F, false);
		syncbtn = GUICreateButton(this, "Synchronize", new Color(0.78F, 0.78F, 0.899F), true, true, SynchronizeTrue);
		syncbtn.button.height = 116F;
		BClr.Paint(syncbtn, BClr.sync, Color.white);

		// select 
		loadSelbtn = GUICreateButton(this, "Load Selection", new Color(0.78F, 0.78F, 0.899F), true, false, LoadSelection);
		loadSelbtn.button.height = 60.0F;
		BClr.Paint(loadSelbtn, BClr.load, Color.white);
		saveSelbtn = GUICreateButton(this, "Save Selection", new Color(0.78F, 0.78F, 0.899F), true, true, SaveSelection);
		saveSelbtn.button.height = 60.0F;
		BClr.Paint(saveSelbtn, BClr.load, Color.white);

		// edit
		loadMorphbtn = GUICreateButton(this, "Load Morphs", new Color(0.78F, 0.78F, 0.899F), true, false, LoadMorphs);
		loadMorphbtn.button.height = 60.0F;
		BClr.Paint(loadMorphbtn, BClr.load, Color.white);
		saveMorphbtn = GUICreateButton(this, "Save Morphs", new Color(0.78F, 0.78F, 0.899F), true, true, SaveMorphs);
		saveMorphbtn.button.height = 60.0F;
		BClr.Paint(saveMorphbtn, BClr.load, Color.white);

		// filter
		GUICreateSpacer(this, false, sep);
		GUICreateSpacer(this, true, sep);
		filterEdit = CreateStringTextbox(false);
		filterSelect = CreateStringTextbox(false);
		filterSelbtn = GUICreateButton(this, "<b>Return to Edit Mode</b>", new Color(0.780f, 0.899f, 0.780f), true, true, Mode2);
		filterSelbtn.button.height = 60.0F;
		BClr.Paint(filterSelbtn, BClr.mode, Color.white);
		filterEditbtn = GUICreateButton(this, "<b>Go to Selection Mode</b>", new Color(0.780f, 0.899f, 0.780f), true, true, Mode1);
		filterEditbtn.button.height = 60.0F;
		BClr.Paint(filterEditbtn, BClr.mode, Color.white);

		// select
		showSelbtn = GUICreateButton(this, "Show Selected", new Color(0.78F, 0.78F, 0.899F), true, false, ShowSelected);
		showSelbtn.button.height = 60.0F;
		BClr.Paint(showSelbtn, BClr.cont, Color.white);
		showAllbtn = GUICreateButton(this, "Show All", new Color(0.78F, 0.78F, 0.899F), true, true, ShowAll);
		showAllbtn.button.height = 60.0F;
		BClr.Paint(showAllbtn, BClr.cont, Color.white);

		selectNonbtn = GUICreateButton(this, "Select None", new Color(0.78F, 0.78F, 0.899F), true, false, SelectNone);
		BClr.Paint(selectNonbtn, BClr.cont, Color.white);
		selectAllbtn = GUICreateButton(this, "Select All", new Color(0.78F, 0.78F, 0.899F), true, true, SelectAll);
		BClr.Paint(selectAllbtn, BClr.cont, Color.white);

		// edit
		origValbtn = GUICreateButton(this, "Original values", new Color(0.78F, 0.78F, 0.899F), true, false, OriginalValues);
		origValbtn.button.height = 60.0F;
		BClr.Paint(origValbtn, BClr.cont, Color.white);
		zeroValbtn = GUICreateButton(this, "Zero values", new Color(0.78F, 0.78F, 0.899F), true, true, ZeroValues);
		zeroValbtn.button.height = 60.0F;
		BClr.Paint(zeroValbtn, BClr.cont, Color.white);
		sp3 = GUICreateSpacer(this, false, sep);
		sp4 = GUICreateSpacer(this, true, sep);

		// ending header
		CreateSpacer(false).height = sep;
		CreateSpacer(true).height = sep;
	}

	#endregion

	#region Controls

	public class LabelEx
	{
		public UIDynamicTextField label;
		public Transform _transform;
		public JSONStorableString value;

		// Constructor (not necessary, but helpful)
		public LabelEx(UIDynamicTextField label, Transform _transform)
		{
			this.label = label;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public LabelEx GUICreateLabel(MVRScript dll, string label, float height, bool rightside)
	{
		JSONStorableString value = new JSONStorableString("", "");
		LabelEx gui = CreateTextField2(value, rightside);
		gui.label.text = label;
		gui.label.height = height;
		gui.value = value;

		return gui;
	}

	public LabelEx CreateTextField2(JSONStorableString jss, bool rightSide = false)
	{
		LabelEx label = null;

		UIDynamicTextField uIDynamicTextField = null;
		if (manager != null && manager.configurableTextFieldPrefab != null && jss.text == null)
		{
			Transform transform = CreateUIElement(manager.configurableTextFieldPrefab.transform, rightSide);
			if (transform != null)
			{
				uIDynamicTextField = transform.GetComponent<UIDynamicTextField>();
				if (uIDynamicTextField != null)
				{
					textFieldToJSONStorableString.Add(uIDynamicTextField, jss);
					jss.dynamicText = uIDynamicTextField;

					label = new LabelEx(uIDynamicTextField, transform);
				}
			}
		}

		return label;
	}

	public class SpacerEx
	{
		public UIDynamic spacer;
		public Transform _transform;

		// Constructor (not necessary, but helpful)
		public SpacerEx(UIDynamic spacer, Transform _transform)
		{
			this.spacer = spacer;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public SpacerEx GUICreateSpacer(MVRScript dll, bool rightside, float height)
	{
		SpacerEx gui = CreateSpacer2(rightside);
		gui.spacer.height = height;

		return gui;
	}

	public SpacerEx CreateSpacer2(bool rightSide = false)
	{
		SpacerEx spacer = null;

		UIDynamic result = null;
		if (manager != null && manager.configurableSpacerPrefab != null)
		{
			Transform transform = CreateUIElement(manager.configurableSpacerPrefab.transform, rightSide);
			if (transform != null)
			{
				result = transform.GetComponent<UIDynamic>();

				spacer = new SpacerEx(result, transform);
			}
		}
		return spacer;
	}

	public class ButtonEx
	{
		public UIDynamicButton button;
		public Transform _transform;

		// Constructor (not necessary, but helpful)
		public ButtonEx(UIDynamicButton button, Transform _transform)
		{
			this.button = button;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public ButtonEx GUICreateButton(MVRScript dll, string label, Color _color, bool useColor, bool rightside, UnityAction callback)
	{
		ButtonEx gui = CreateButton2(label, rightside);
		if (useColor) { gui.button.buttonColor = _color; }
		gui.button.button.onClick.AddListener(callback);

		return gui;
	}

	public ButtonEx CreateButton2(string label, bool rightSide = false)
	{
		ButtonEx button = null;

		UIDynamicButton uIDynamicButton = null;
		if (manager != null && manager.configurableButtonPrefab != null)
		{
			Transform transform = CreateUIElement(manager.configurableButtonPrefab.transform, rightSide);
			if (transform != null)
			{
				uIDynamicButton = transform.GetComponent<UIDynamicButton>();
				if (uIDynamicButton != null)
				{
					uIDynamicButton.label = label;

					button = new ButtonEx(uIDynamicButton, transform);
				}
			}
		}

		return button;
	}

	public class ToggleEx
	{
		public UIDynamicToggle toggle;
		public JSONStorableBool value;
		public Transform _transform;

		// Constructor (not necessary, but helpful)
		public ToggleEx(UIDynamicToggle toggle, Transform _transform)
		{
			this.toggle = toggle;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public ToggleEx GUICreateCheckbox(MVRScript dll, string name, string label, bool startvalue, bool rightside, JSONStorableBool.SetJSONBoolCallback callback)
	{
		JSONStorableBool gui = new JSONStorableBool(name, startvalue, callback);
		ToggleEx Toggle = CreateToggle2(gui, rightside);
		Toggle.toggle.label = label;
		dll.RegisterBool(gui);
		Toggle.value = gui;

		return Toggle;
	}

	public ToggleEx CreateToggle2(JSONStorableBool jsb, bool rightSide = false)
	{
		ToggleEx toggle = null;

		UIDynamicToggle uIDynamicToggle = null;
		if (manager != null && manager.configurableTogglePrefab != null && jsb.toggle == null)
		{
			Transform transform = CreateUIElement(manager.configurableTogglePrefab.transform, rightSide);
			if (transform != null)
			{
				uIDynamicToggle = transform.GetComponent<UIDynamicToggle>();
				if (uIDynamicToggle != null)
				{
					toggleToJSONStorableBool.Add(uIDynamicToggle, jsb);
					uIDynamicToggle.label = jsb.name;
					jsb.toggle = uIDynamicToggle.toggle;

					toggle = new ToggleEx(uIDynamicToggle, transform);
				}
			}
		}

		return toggle;
	}

	public class SliderEx
	{
		public UIDynamicSlider slider;
		public JSONStorableFloat value;
		public Transform _transform;

		// Constructor (not necessary, but helpful)
		public SliderEx(UIDynamicSlider slider, Transform _transform)
		{
			this.slider = slider;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public SliderEx GUICreateSlider(MVRScript dll, string name, string label, float startvalue, bool rightside, float minvalue, float maxvalue, UnityAction<float> _event)
	{
		JSONStorableFloat gui = new JSONStorableFloat(name, startvalue, minvalue, maxvalue);
		SliderEx Slider = CreateSlider2(gui, rightside);
		Slider.slider.label = label;
		Slider.value = gui;
		dll.RegisterFloat(gui);
		Slider.value = gui;
		Slider.slider.slider.onValueChanged.AddListener(_event);

		return Slider;
	}

	public SliderEx CreateSlider2(JSONStorableFloat jsf, bool rightSide = false)
	{
		SliderEx slider = null;

		UIDynamicSlider uIDynamicSlider = null;
		if (manager != null && manager.configurableSliderPrefab != null && jsf.slider == null)
		{
			Transform transform = CreateUIElement(manager.configurableSliderPrefab.transform, rightSide);
			if (transform != null)
			{
				uIDynamicSlider = transform.GetComponent<UIDynamicSlider>();
				if (uIDynamicSlider != null)
				{
					uIDynamicSlider.Configure(jsf.name, jsf.min, jsf.max, jsf.val, jsf.constrained, "F2", true, !jsf.constrained);
					jsf.slider = uIDynamicSlider.slider;
					sliderToJSONStorableFloat.Add(uIDynamicSlider, jsf);

					slider = new SliderEx(uIDynamicSlider, transform);
				}
			}
		}
		return slider;
	}

	public class PopupEx
	{
		public UIDynamicPopup popup;
		public JSONStorableStringChooser value;
		public Transform _transform;

		// Constructor (not necessary, but helpful)
		public PopupEx(UIDynamicPopup popup, Transform _transform)
		{
			this.popup = popup;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public PopupEx GUICreatePopup(MVRScript dll, string name, string label, bool rightside, UIPopup.OnOpenPopup popup, JSONStorableStringChooser.SetStringCallback callback)
	{
		JSONStorableStringChooser gui = new JSONStorableStringChooser(name, null, "", "", callback);
		dll.RegisterStringChooser(gui);
		PopupEx dp = CreateFilterablePopup2(gui, rightside);
		dp.popup.popupPanelHeight = 500.0F;
		dp.popup.label = label;
		dp.popup.popup.onOpenPopupHandlers = popup;
		dp.value = gui;
		return dp;
	}

	public PopupEx CreateFilterablePopup2(JSONStorableStringChooser jsc, bool rightSide = false)
	{
		PopupEx popup = null;

		UIDynamicPopup uIDynamicPopup = null;
		if (manager != null && manager.configurableFilterablePopupPrefab != null && jsc.popup == null)
		{
			Transform transform = CreateUIElement(manager.configurableFilterablePopupPrefab.transform, rightSide);
			if (transform != null)
			{
				uIDynamicPopup = transform.GetComponent<UIDynamicPopup>();
				if (uIDynamicPopup != null)
				{
					popupToJSONStorableStringChooser.Add(uIDynamicPopup, jsc);
					uIDynamicPopup.label = jsc.name;
					jsc.popup = uIDynamicPopup.popup;

					popup = new PopupEx(uIDynamicPopup, transform);
				}
			}
		}

		return popup;
	}

	public LabelEx CreateStringTextbox(bool rhs)
	{
		JSONStorableString json = new JSONStorableString("", "");
		LabelEx textField = CreateTextField2(json, rhs);
		SetTextFieldHeight(textField.label, 60);
		InputField input = textField.label.gameObject.AddComponent<InputField>();
		input.textComponent = textField.label.UItext;
		input.textComponent.fontSize = 32; // 32
		textField.label.backgroundColor = Color.white;
		json.inputField = input;
		textField.value = json;
		input.onValueChanged.AddListener((string value) => {
			json.val = value;
		});
		return textField;
	}

	public void SetTextFieldHeight(UIDynamicTextField textField, int height)
	{
		LayoutElement component = textField.GetComponent<LayoutElement>();
		if (component != null)
		{
			component.minHeight = height;
			component.preferredHeight = height;
		}
		textField.height = height;
	}

	#endregion

	#region Install

	private void Install()
	{
		string newDir = _saveFolder;

		if (FileManagerSecure.DirectoryExists(newDir) == false)
			FileManagerSecure.CreateDirectory(newDir);

		newDir = _saveFolderSel;

		if (FileManagerSecure.DirectoryExists(newDir) == false)
			FileManagerSecure.CreateDirectory(newDir);
	}

	#endregion

	#region Synchronize

	private void updateStatus()
	{
		int count = allmorphs.Count; //  - actmorphs.Count;
		status.label.text = "<b>Available</b>: " + count.ToString() + " morphs\n<b>Active</b>: " + slotList.Count.ToString() + " morphs\n<b>Selected</b>: <b>" + selectedCount.ToString() + " morphs</b>";
	}

	private void SynchronizeTrue()
	{
		Synchronize(true);
	}

	private void Synchronize(bool syncing)
	{
		GetActiveMorphs();
		BuildGui(syncing);
		filterSelect.value.val = "Selected";
		filterEdit.value.val = "";
		SelectMode(editing);
		updateStatus();
	}

	#endregion

	#region Build Morphs

	private void BuildGui(bool syncing)
	{
		sideToggle = true;
		sideSlider = true;
		string lastLetter = actmorphs[0].displayName;
		char a = lastLetter[0];

		foreach (DAZMorph morph in actmorphs)
		{
			Slot myslot = null;

			if (firstTime == false)
			{
				myslot = AddSlot(morph, false);
			}
			else
			{
				string display = morph.displayName;

				if (display[0] == a)
				{
					myslot = AddSlot(morph, false);
				}
				else
				{
					myslot = AddSlot(morph, true);
				}

				a = display[0];
			}

			if (syncing == true)
            {
				if (myslot != null)
                {
					myslot.SetSelected(true);
					myslot.origValue = 0F;
					if (editing) { selectedCount++; }
				}
			}
		}
	}

	private Slot AddSlot(DAZMorph morph, bool isbold)
	{
		bool exists = false;
		Slot myslot = null;

		foreach (Slot slot in slotList)
		{
			if (slot.morph.displayName.Equals(morph.displayName))
			{
				exists = true;
				break;
			}
		}

		if (exists == false)
		{
			myslot = new Slot(morph, this, isbold);
			slotList.Add(myslot);
		}

		return myslot;
	}

	#endregion

	#region Update

	public void Update()
	{
		if (Input.GetKeyUp(KeyCode.Return))
		{
			if (editing)
            {
				FilterEdit();
            }
			else
            {
				FilterSelect();
            }
		}
	}
	
	#endregion

} // End class
